/************************************************************************
 *
 *                    CalC (Calcium Calculator)
 *                         Victor Matveev
 *
 *                            box.cpp
 *
 *               Geometry definition classes:
 * 
 *               Generalized box class BoxObj
 *               Box as an obstacle: class ObstBoxObj
 *               Set of boxes and obstacles: RegionObj
 *
 ************************************************************************/

#include "stdafx.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdarg.h>
#include "syntax.h"
#include "grid.h"
#include "box.h"

class GridObj *BoxObj::Grid = 0;
class GridObj *RegionObj::Grid = 0;


//***************************************************************
//                I M P L E M E N T A T I O N S
//***************************************************************

int BoxObj::inside(int ix, int iy, int iz) {
    
  if (ix < 0 || ix >= Grid->xsize) return 0;
  bool b = (Grid->xcoord[ix] < xmax + 0.5 * Grid->xgrid[ix - ( ix > 0 ? 1 : 0)]) && 
           (Grid->xcoord[ix] > xmin - 0.5 * Grid->xgrid[ix]);

  if ( DIMENSIONALITY > 1) {
    if (iy < 0 || iy >= Grid->ysize) return 0;
    b = b && (Grid->ycoord[iy] < ymax + 0.5 * Grid->ygrid[iy - ( iy > 0 ? 1 : 0) ]) && 
             (Grid->ycoord[iy] > ymin - 0.5 * Grid->ygrid[iy]);
  }

  if (DIMENSIONALITY > 2) {
     if (iz < 0 || iz >= Grid->zsize) return 0;
     b = b && (Grid->zcoord[iz] < zmax + 0.5 * Grid->zgrid[iz - ( iz > 0 ? 1 : 0) ]) && 
	      (Grid->zcoord[iz] > zmin - 0.5 * Grid->zgrid[iz]);
  }

  if (b) return _INSIDE_;  else return 0;
}

//***************************************************************

int ObstBoxObj::inside(int ix, int iy, int iz)
  {    
  if (ix < 0 || ix >= Grid->xsize) return _INSIDE_;
  bool b = (Grid->xcoord[ix] < xmax - 0.5 * Grid->xgrid[ix]) && 
           (Grid->xcoord[ix] > xmin + 0.5 * Grid->xgrid[ix - ( ix > 0 ? 1 : 0)]);
  if (DIMENSIONALITY > 1) {
     if (iy < 0 || iy >= Grid->ysize) return _INSIDE_;
     b = b && (Grid->ycoord[iy] < ymax - 0.5 * Grid->ygrid[iy]) && 
	      (Grid->ycoord[iy] > ymin + 0.5 * Grid->ygrid[iy - ( iy > 0 ? 1 : 0) ]);
  }
  if (DIMENSIONALITY > 2) {
     if (iz < 0 || iz >= Grid->zsize) return _INSIDE_;
     b = b && (Grid->zcoord[iz] < zmax - 0.5 * Grid->zgrid[iz]) && 
      	      (Grid->zcoord[iz] > zmin + 0.5 * Grid->zgrid[iz - ( iz > 0 ? 1 : 0) ]);
  }

  if (b) return 0; else return _INSIDE_;
  }

//***************************************************************

int BoxObj::point_type(int ix, int iy, int iz)
   {
   int  surface;

   if ( (surface = inside(ix, iy, iz)) > 0 )
      {
      if (! inside(ix - 1, iy, iz) ) surface += SURF_XMIN;
      if (! inside(ix + 1, iy, iz) ) surface += SURF_XMAX;
      if (DIMENSIONALITY == 1) return surface;
      if (! inside(ix, iy - 1, iz) ) surface += SURF_YMIN;
      if (! inside(ix, iy + 1, iz) ) surface += SURF_YMAX;
      if (DIMENSIONALITY == 2) return surface;
      if (! inside(ix, iy, iz - 1) ) surface += SURF_ZMIN;
      if (! inside(ix, iy, iz + 1) ) surface += SURF_ZMAX;
      }

    return surface;
    }


//***************************************************************
//*   C L A S S   R E G I O N  I M P L E M E N T A T I O N S
//***************************************************************

RegionObj::RegionObj(TokenString &pars)
{

 if (VERBOSE) { fprintf(stderr,"\n#### Building the geometry:"); fflush(stderr); }

 encl_num = pars.token_count("volume");
 if (!encl_num) globalError(StrCpy("No diffusion volume definitions in script file"));

 initialize( encl_num, pars.token_count("obstacle") );

 int i;
 for (i=1; i<=encl_num; i++)  add_enclosure(pars);
 if (obst_num)  for (i=1; i<=obst_num; i++)  add_obstacle(pars);
}

//***************************************************************

 void RegionObj::initialize(int vNum, int oNum) {  
   encl_cnt = obst_cnt = 0;
   encl_num = vNum; obst_num = oNum;
   total = encl_num + obst_num;
   enclosures = new BoxObj[encl_num];
   obstacles = new ObstBoxObj[obst_num];
 }

//***************************************************************

RegionObj::~RegionObj()
{
  delete [] enclosures;
  delete [] obstacles;
}

//***************************************************************
// point_type bit fields are:
// bit # is:    ...8     |   7    |  6   |  5   |  4   |  3   |  2   |  1  
// meaning:   box number | inside | zmax | zmin | ymax | ymin | xmax | xmin
//***************************************************************


int RegionObj::point_type(int ix, int iy, int iz, int extObstNum, class ObstBoxObj *OBO)
{
int    ptype, boxid = 0;
int    encl_bnd = SURF_MASK;
int    obst_bnd = SURF_MASK;  
int    i, j, inside = 0;

    for (i = 0; i < encl_num; i++)
      {  
      ptype = enclosures[i].point_type(ix, iy, iz); 
      if (ptype & _INSIDE_)    
        {                                      //  Point is inside if it is
	inside = _INSIDE_;                     // inside any of the enclosures.
        encl_bnd &= ptype;                     //  Inside point is on a boundary only
        if (ptype  & SURF_MASK )  boxid = i;   // if it belongs to the same boundary
	}                                      // of all the enclosures containing it
      }                                       
    if (encl_bnd == SURF_MASK)  encl_bnd = 0;  

    for (j = 0; j < obst_num; j++)
      {  
      ptype = obstacles[j].point_type(ix, iy, iz);
      if (ptype == 0) inside = 0;              // Point is outside if it belongs
        else                                   // to any of the obstacles.
          if (ptype  & SURF_MASK )  
            { 
            obst_bnd &= ptype;                 // Inside point is on a boundary only
            boxid = j + encl_num;              // if it belongs to the same boundary
            if (obst_bnd == 0) inside = 0;     // of all the obstacles
            }                                  // Point is outside if it belongs to
      }                                        // boundaries of different orientations
    

    if (extObstNum)
    for (j = 0; j < extObstNum; j++)
      {  
      ptype = OBO[j].point_type(ix, iy, iz);
      if (ptype == 0) inside = 0;              // Point is outside if it belongs
        else                                   // to any of the obstacles.
          if (ptype  & SURF_MASK )  
            { 
            obst_bnd &= ptype;                 // Inside point is on a boundary only
            boxid = j + encl_num;              // if it belongs to the same boundary
            if (obst_bnd == 0) inside = 0;     // of all the obstacles
            }                                  // Point is outside if it belongs to
      }                                        // boundaries of different orientations

    if (obst_bnd == SURF_MASK)  obst_bnd = 0;  

    if ( inside == 0 ) return 0;

    int bnd = encl_bnd + obst_bnd;
 
    if (bnd == 0) return inside;

    if ( ((bnd & SURF_XMIN) && (bnd & SURF_XMAX)) ||
         ((bnd & SURF_YMIN) && (bnd & SURF_YMAX)) ||
         ((bnd & SURF_ZMIN) && (bnd & SURF_ZMAX)) )
      
    throw makeMessage("Ambiguity: two surfaces touch. Please enlarge the obstacle dimensions, extending it beyond the domain boundary, to avoid surface ambiguity");

    else  return  _INSIDE_ + (boxid << 7) + bnd;
   
}


//***************************************************************

int RegionObj::point_type(long ind, int extObstNum, class ObstBoxObj *OBO)
{
int iz = ind / Grid->xysize;
ind -= iz * Grid->xysize;
int iy = ind / Grid->xsize;
int ix = ind - iy * Grid->xsize;
return point_type(ix, iy, iz, extObstNum, OBO);
}

//***************************************************************

void RegionObj::add_enclosure(double x0, double x1, double y0, double y1,
                              double z0, double z1)
{
 if (encl_cnt >= encl_num) 
   throw makeMessage("RegionObj::add_enclosure: encl_cnt=%d >= encl_num=%d\n",encl_num, encl_cnt);

 if (encl_cnt == 0) 
   { xmin = x0; xmax = x1; ymin = y0; ymax = y1; zmin = z0; zmax = z1; }
 else
   {
   if (x0 < xmin) xmin = x0;  if (x1 > xmax) xmax = x1;
   if (y0 < ymin) ymin = y0;  if (y1 > ymax) ymax = y1;
   if (z0 < zmin) zmin = z0;  if (z1 > zmax) zmax = z1;
   }

 if (VERBOSE) { fprintf(stderr,"   adding enclosure #%d:", encl_cnt+1); fflush(stderr); }

 enclosures[encl_cnt] = BoxObj(x0, x1, y0, y1, z0, z1);

 if (VERBOSE) {
   fprintf(stderr," [%g,%g]", x0, x1);
   if (DIMENSIONALITY > 1) fprintf(stderr, " x [%g,%g]", y0, y1);
   if (DIMENSIONALITY > 2) fprintf(stderr, " x [%g,%g]", z0, z1);
   fprintf(stderr, "\n");
 }

 if ( x0 > x1 || y0 > y1 || z0 > z1 )  throw StrCpy("Check enclosure volume dimensions");

 encl_cnt++;
}

//***************************************************************

void RegionObj::add_enclosure(TokenString &pars)
{
  double x0, x1, y0, y1, z0, z1;
  long   pos;

  try {
    pos = pars.token_index("volume", encl_cnt+1);
    if (DIMENSIONALITY == 1)
      {
      pars.trail_pars(pos, 'd', &x0, 'd', &x1, 'E');
      add_enclosure(x0, x1, 0.0, 0.0, 0.0, 0.0);
      }
    else  if (DIMENSIONALITY == 2)
      {
      pars.trail_pars(pos, 'd', &x0, 'd', &x1, 'd', &y0, 'd', &y1, 'E');
      add_enclosure(x0, x1, y0, y1, 0.0, 0.0);
      }
    else
      {
      pars.trail_pars(pos, 'd', &x0, 'd', &x1, 'd', &y0, 'd', &y1, 'd', &z0, 'd', &z1, 'E');    
      add_enclosure(x0, x1, y0, y1, z0, z1);
      }
  }
  catch(char *message) 
    { pars.errorMessage( pos + 1, message); }
}

//***************************************************************

void RegionObj::add_obstacle(double x0, double x1, double y0, double y1, 
                             double z0, double z1)
{
 if (obst_cnt >= obst_num)  
   throw makeMessage("RegionObj::add_obstacle: obst_cnt=%d >= obst_num=%d\n",obst_num, obst_cnt);

 if (VERBOSE) { fprintf(stderr,"   adding obstacle #%d:", obst_cnt+1); fflush(stderr); }

 obstacles[obst_cnt] = ObstBoxObj(x0, x1, y0, y1, z0, z1);

 if (VERBOSE)  {
   fprintf(stderr," [%g,%g]", x0, x1);
   if (DIMENSIONALITY > 1) fprintf(stderr, " x [%g,%g]", y0, y1);
   if (DIMENSIONALITY > 2) fprintf(stderr, " x [%g,%g]", z0, z1);
   fprintf(stderr, "\n");
 }

 if ( x0 > x1 || y0 > y1 || z0 > z1 )   throw StrCpy("Check diffusion obstacle dimensions"); 

 obst_cnt++;
}

//***************************************************************

void RegionObj::add_obstacle(TokenString &pars)
{
  double x0, x1, y0, y1, z0, z1;
  long   pos;

  try {
    pos = pars.token_index("obstacle", obst_cnt+1);
    if (DIMENSIONALITY == 1)
      {
      pars.trail_pars(pos, 'd', &x0, 'd', &x1, 'E');
      add_obstacle(x0, x1, 0.0, 0.0, 0.0, 0.0);
      }
    else  if (DIMENSIONALITY == 2)
      {
      pars.trail_pars(pos, 'd', &x0, 'd', &x1, 'd', &y0, 'd', &y1, 'E');
      add_obstacle(x0, x1, y0, y1, 0.0, 0.0);
      }
    else  // 3 spatial dimensions
      {
      pars.trail_pars(pos, 'd', &x0, 'd', &x1, 'd', &y0, 'd', &y1, 'd', &z0, 'd', &z1, 'E');    
      add_obstacle(x0, x1, y0, y1, z0, z1);
      }
    }
  catch(char *message)
    { pars.errorMessage( pos + 1, message); }
}

//***************************************************************

void print_type(int x)
{
if (x == 0) fprintf(stderr,"[OUT] ");
else
  {
  if (x & _INSIDE_) fprintf(stderr,"[IN");
  if (!(x & SURF_MASK)) fprintf(stderr,"] ");
    else
      {
      if (x & SURF_XMIN) fprintf(stderr," XMIN");
      if (x & SURF_XMAX) fprintf(stderr," XMAX");
      if (x & SURF_YMIN) fprintf(stderr," YMIN");
      if (x & SURF_YMAX) fprintf(stderr," YMAX");
      if (x & SURF_ZMIN) fprintf(stderr," ZMIN");
      if (x & SURF_ZMAX) fprintf(stderr," ZMAX");
      fprintf(stderr," %d] ",x>>9);
      }
  }
}
#ifdef CALC_DEBUG
#endif
//***************************************************************
//                       T H E   E N D
//***************************************************************

